--[[
StorageExtension.lua
	
Author: 	Ifko[nator]
Date:		13.06.2021
Version:	1.2

History:	v1.0 @12.09.2020 - initial implementation in FS 19
			--------------------------------------------------
			v1.1 @05.06.2021 - some script optimation
			--------------------------------------------------
			v1.2 @13.06.2021 - minior adjustments
]]

StorageExtension = {};

function StorageExtension:StorageLoad(superFunc, id, xmlFile, key)
	if RealisticSeederUtil == nil then
		return superFunc(id, xmlFile, key);
	end;

	self.rootNode = id;
	self.capacityPerFillType = getXMLFloat(xmlFile, key .. "#capacityPerFillType") or 100000;
	self.costsPerFillLevelAndDay = getXMLFloat(xmlFile, key .. "#costsPerFillLevelAndDay") or 0;
	self.fillTypes = {};
	self.fillLevels = {};
    self.sortedFillTypes = {};
    
	local fillTypeCategories = getXMLString(xmlFile, key .. "#fillTypeCategories");
	local fillTypeNames = getXMLString(xmlFile, key .. "#fillTypes");
	local fillTypes = nil;

	if fillTypeCategories ~= nil and fillTypeNames == nil then
		fillTypes = g_fillTypeManager:getFillTypesByCategoryNames(fillTypeCategories, "Warning: '" .. tostring(key) .. "' has invalid fillTypeCategory '%s'.");
	elseif fillTypeCategories == nil and fillTypeNames ~= nil then
		fillTypes = g_fillTypeManager:getFillTypesByNames(fillTypeNames, "Warning: '" .. tostring(key) .. "' has invalid fillType '%s'.");
	else
		print("Warning: '" .. tostring(key) .. "' a 'Storage' entry needs either the 'fillTypeCategories' or 'fillTypes' attribute.");

		return false;
	end;

	for _, fillType in pairs(fillTypes) do
		if not self.fillTypes[fillType] then    
            self.fillTypes[fillType] = true;
        end;
	end;

    local fillTypes = RealisticSeederUtil.getSeedFillTypes();
    
    if fillTypes ~= nil then
        for _, fillType in pairs(fillTypes) do
            --## check if selected fruit is available in current map
	    	if RealisticSeederUtil.getCanFruitBePlanted(fillType) and self.fillTypes[g_fillTypeManager:getFillTypeByName("wheat")] then
                if not self.fillTypes[fillType] then    
                    self.fillTypes[fillType] = true;
                end;
            end;
        end;
    end;

	for fillType, _ in pairs(self.fillTypes) do
		table.insert(self.sortedFillTypes, fillType);

		self.fillLevels[fillType] = 0;
	end;

	table.sort(self.sortedFillTypes);

	self.storageDirtyFlag = self:getNextDirtyFlag();

	g_messageCenter:subscribe(MessageType.FARM_DELETED, self.farmDestroyed, self);

	return true;
end;

Storage.load = Utils.overwrittenFunction(Storage.load, StorageExtension.StorageLoad);

function StorageExtension:setFillLevels(superFunc, fillLevels, capacity, hasInfiniteCapacity)
	self.fillLevels = fillLevels;
	self.capacity = capacity;
	self.fillTypeMapping = {};

	local fillTypesTable = {};
	local selectedId = 1;
	local numFillLevels = 1;

	for fillTypeIndex, _ in pairs(fillLevels) do
		local fillType = g_fillTypeManager:getFillTypeByIndex(fillTypeIndex);
		local level = Utils.getNoNil(fillLevels[fillTypeIndex], 0);
		local name = fillType.title;
		local maxTextLenght = 20;
		local originalTextLenght = string.len(name);

		if originalTextLenght > maxTextLenght then
			name = string.sub(name, 1, maxTextLenght) .. "..";
		end;

		if not hasInfiniteCapacity then
			name = string.format("%s\n%s", name, g_i18n:formatFluid(level));
		end;

		table.insert(fillTypesTable, name);
		table.insert(self.fillTypeMapping, fillTypeIndex);

		if fillTypeIndex == self.lastSelectedFillType then
			selectedId = numFillLevels;
		end;

		numFillLevels = numFillLevels + 1;
	end;

	self.fillTypesElement:setTexts(fillTypesTable);
	self.fillTypesElement:setState(selectedId, true);
end;

SiloDialog.setFillLevels = Utils.overwrittenFunction(SiloDialog.setFillLevels, StorageExtension.setFillLevels);

addModEventListener(StorageExtension);